#include "streaminfo.h"

StreamInfo::StreamInfo(const string &streamPath)
    :streamPath_(streamPath)
{
    extractedPointcloud_.reset(new PointCloudXYZI);
    load();
}

void StreamInfo::load()
{
    ifstream infile(streamPath_, ios::binary);
    if(!infile)
    {
        LOG(WARNING) << streamPath_ << " open error!";
        return ;
    }
    char buf[2];
    infile.read(buf, 2);
    headerLength_ = ((uint16_t)buf[1] << 8) + (uint16_t)buf[0];
    LOG(INFO) << "headerLength: " << headerLength_;

    char* headerData = (char*)malloc(headerLength_);
    infile.read(headerData, headerLength_);
    header_.ParseFromArray(headerData, headerLength_);
    free(headerData);
    float grid_relolution = header_.grid_relolution();
    float block_relolution = header_.block_relolution();
    offset_.x() = header_.offset_x();
    offset_.y() = header_.offset_y();
    LOG(INFO) << "grid_relolution: " << grid_relolution;
    LOG(INFO) << "block_relolution: " << block_relolution;

    proj_.reset(new LocalCartesian(header_.latitude(),
                                   header_.longtitude(),
                                   header_.height()));

    uint16_t blockIndexLength = header_.block_index_length();
    LOG(INFO) << "blockIndexLength: " << blockIndexLength;
    char* blockIndexData = (char*)malloc(blockIndexLength);
    infile.read(blockIndexData, blockIndexLength);
    blockIndex_.ParseFromArray(blockIndexData, blockIndexLength);
    free(blockIndexData);
    size_t block_describe_size = blockIndex_.block_describe_size();
    for(size_t i = 0; i < block_describe_size; i++){
        BlockDescribe blockDescribe = blockIndex_.block_describe(i);
        Internal::BlockDescribe internalBlockDescribe;
        internalBlockDescribe.blockId = blockDescribe.block_id();
        internalBlockDescribe.blockOffset = blockDescribe.block_info_offset();
        if(i != 0){
            blockDescribeVec_.back().blockSize =
                    internalBlockDescribe.blockOffset - blockDescribeVec_.back().blockOffset;
        }
        blockDescribeVec_.push_back(internalBlockDescribe);
    }
    blockDescribeVec_.back().blockSize =
            header_.block_info_length() - blockDescribeVec_.back().blockOffset;

    LOG(INFO) << "block_describe_size: " << block_describe_size;

    infile.clear();
    infile.seekg(0, infile.end);
    uint32_t totalSize = infile.tellg();
    LOG(INFO) << "infile.tellg(): " << infile.tellg();
    infile.clear();
    infile.seekg(2 + headerLength_ +
                 header_.block_index_length() + header_.block_info_length());

    uint32_t distributionSize = totalSize - 2 - headerLength_ -
            header_.block_index_length() - header_.block_info_length();

    char* distributionBuf = (char*)malloc(distributionSize);

    LOG(INFO) << "distributionSize: " << distributionSize;
    infile.read(distributionBuf, distributionSize);
    infile.close();

    Distributions_.ParseFromArray(distributionBuf, distributionSize);
    free(distributionBuf);

    LOG(INFO) << "Distributions_.grid_distribution_size(): " << Distributions_.grid_distribution_size();
    for(size_t i = 0; i < blockDescribeVec_.size(); i++){
        auto blockDescribe = blockDescribeVec_[i];
        parseOneBlock(blockDescribe);
//        if(extractedPointcloud_->size()){
//            pcl::io::savePCDFileBinary(
//                        "/home/juefx/temp/" + to_string(i) + "_block.pcd", *extractedPointcloud_);
//        }
//        extractedPointcloud_->clear();
    }
    LOG(INFO) << "extractedPointcloud_.size(): " << extractedPointcloud_->size();
    if(extractedPointcloud_->size()){
        pcl::io::savePCDFileBinary("/home/juefx/extracted.pcd", *extractedPointcloud_);
    }

//    voxel_.calculateHits(extractedPointcloud_);
}

void StreamInfo::parseOneBlock(
        const Internal::BlockDescribe &internalBlockDescribe)
{
    ifstream infile(streamPath_, ios::binary);
    if(!infile)
    {
        LOG(WARNING) << streamPath_ << " open error!";
        return ;
    }

    char* blockInfoData = (char*)malloc(internalBlockDescribe.blockSize);
    uint32_t infoOffset = 2 + headerLength_ + header_.block_index_length() + internalBlockDescribe.blockOffset;
    infile.clear();
    infile.seekg(infoOffset);
    infile.read(blockInfoData, internalBlockDescribe.blockSize);
    BlockInfo blockInfo;
    blockInfo.ParseFromArray(blockInfoData, internalBlockDescribe.blockSize);
    free(blockInfoData);
    infile.close();

//    LOG(INFO) << "blockIndexX: " << blockIndexX;
//    LOG(INFO) << "blockIndexY: " << blockIndexY;
//    LOG(INFO) << "blockCenterX: " << blockCenterX;
//    LOG(INFO) << "blockCenterY: " << blockCenterY;
    if(blockInfo.bit_map_size()){
        parseBitmapBlock(blockInfo, internalBlockDescribe);
    }else{
        parseIndexBlock(blockInfo, internalBlockDescribe);
    }

}

vector<uint32_t> StreamInfo::parseBitMap(
        const ProtoRepeatedUint32 &bitmap,
        const uint32_t size)
{
    vector<uint32_t> activeGridIndexes;
    uint32_t activeGridIndex = 0;
    for(size_t i = 0; i < size; i++){
        uint8_t byte = bitmap.Get(i);
        for(size_t byteCnt = 0; byteCnt < 8; byteCnt++){
            if((byte >> byteCnt) & 1){
                activeGridIndexes.push_back(activeGridIndex);
            }
            activeGridIndex++;
        }
    }
//    LOG(INFO) << "activeGridIndexes.size(): " << activeGridIndexes.size();
    return activeGridIndexes;
}

void StreamInfo::Forward(double lat, double lon, double h, double &x, double &y, double &z)
{
    proj_->Forward(lat, lon, h, x, y, z);
}

void StreamInfo::parseBitmapBlock(
        const BlockInfo& blockInfo,
        const Internal::BlockDescribe &internalBlockDescribe)
{
    float gridResolution = blockInfo.grid_resolution();
    int32_t blockId = internalBlockDescribe.blockId;
    int32_t blockIndexX = blockId / (int32_t)(1000 / BlockResolution);
    int32_t blockIndexY = blockId % (int32_t)(1000 / BlockResolution);
    float blockCenterX = ((float)blockIndexX + 0.5) * BlockResolution;
    float blockCenterY = ((float)blockIndexY + 0.5) * BlockResolution;

    vector<uint32_t> activeGridIndexes =
            parseBitMap(blockInfo.bit_map(), blockInfo.bit_map_size());

    for(size_t i = 0; i < activeGridIndexes.size(); i++){
        auto index = activeGridIndexes[i];
        auto grid_feature_index = blockInfo.grid_feature_index(i);
        GridDistribution gridDistribution = Distributions_.grid_distribution(grid_feature_index);
        uint32_t indexX = index / (uint32_t)(BlockResolution / gridResolution);
        uint32_t indexY = index % (uint32_t)(BlockResolution / gridResolution);
        size_t oneDistributionCnt = gridDistribution.one_distribution_size();

        float centerX = ((float)indexX + 0.5) * gridResolution - BlockResolution / 2;
        float centerY = ((float)indexY + 0.5) * gridResolution - BlockResolution / 2;
        for(size_t i = 0; i < oneDistributionCnt; i++){
            OneDistribution oneDistribution = gridDistribution.one_distribution(i);
            for(int32_t j = 0;
                j < oneDistribution.z_delta_leval();
                j++){
                int32_t indexZ = j + oneDistribution.floor();
                pcl::PointXYZI point;
                point.x = centerX + blockCenterX + offset_.x();
                point.y = centerY + blockCenterY + offset_.y();
                point.z = ((float)indexZ + 0.5) * ZResolution;
                point.intensity = oneDistribution.intensity() * 8;
                extractedPointcloud_->push_back(point);
            }
        }
    }
}

void StreamInfo::parseIndexBlock(
        const BlockInfo& blockInfo,
        const Internal::BlockDescribe &internalBlockDescribe)
{
    float gridResolution = blockInfo.grid_resolution();
    int32_t blockId = internalBlockDescribe.blockId;
    int32_t blockIndexX = blockId / (1000 / (int32_t)BlockResolution);
    int32_t blockIndexY = blockId % (1000 / (int32_t)BlockResolution);
    float blockCenterX = ((float)blockIndexX + 0.5) * BlockResolution;
    float blockCenterY = ((float)blockIndexY + 0.5) * BlockResolution;

    for(size_t i = 0; i < blockInfo.grid_feature_index_size(); i++){
        auto grid_feature_index = blockInfo.grid_feature_index(i);
        if(grid_feature_index == 0){
            continue;
        }
        grid_feature_index--;

        GridDistribution gridDistribution = Distributions_.grid_distribution(grid_feature_index);
        uint32_t indexX = i / (BlockResolution / gridResolution);
        uint32_t indexY = i % (uint32_t)(BlockResolution / gridResolution);
        size_t oneDistributionCnt = gridDistribution.one_distribution_size();

        float centerX = ((float)indexX + 0.5) * gridResolution - BlockResolution / 2;
        float centerY = ((float)indexY + 0.5) * gridResolution - BlockResolution / 2;
        for(size_t i = 0; i < oneDistributionCnt; i++){
            OneDistribution oneDistribution = gridDistribution.one_distribution(i);
            for(int32_t j = 0;
                j < oneDistribution.z_delta_leval();
                j++){
                int32_t indexZ = j + oneDistribution.floor();
                pcl::PointXYZI point;
                point.x = centerX + blockCenterX + offset_.x();
                point.y = centerY + blockCenterY + offset_.y();
                point.z = ((float)indexZ + 0.5) * ZResolution;
                point.intensity = oneDistribution.intensity() * 8;
                extractedPointcloud_->push_back(point);
            }
        }
    }
}
