﻿#ifndef BASIC_TYPES_H
#define BASIC_TYPES_H
#pragma once
#include <boost/shared_ptr.hpp>
#include <eigen3/Eigen/Geometry>
#include "glog/logging.h"
#include <map>
#include <iomanip>
#include <pcl/io/pcd_io.h>
#include <thread>
#include <boost/make_shared.hpp>

using namespace std;
using namespace Eigen;
#define MAX_THREAD_CNT (48)

typedef boost::shared_ptr<boost::thread> ThreadPtr;

/////////// dispatch /////////////

const float BlockResolution = 18.;
const float GridResolution6cm = 0.06;
const float GridResolution24cm = 0.24;
const float ZResolution = 0.03;

enum GridType{
    blockType,
    gridType
};

struct OneDistribution_{
  int32_t floor = 1;
  uint32_t z_delta_level = 2;
  uint32_t intensity = 3;
};

struct GridDistribution_{
    vector<OneDistribution_> oneDistributions;
    int64_t getScore(){
        int64_t score = 0;
        if(oneDistributions.size() == 0){
            return 0;
        }
        score += oneDistributions.size() * 1e12;
        for(const OneDistribution_& oneDistribution : oneDistributions){
            score += oneDistribution.z_delta_level * 1e8;
            score += oneDistribution.floor * 1e4;
            score += oneDistribution.intensity;
        }
        return score;
    }
};

struct BlockDistribution_{
    vector<GridDistribution_> gridDistributions;
};

struct IndexDistribution{
    GridDistribution_ gridDistribution_;
    map<int32_t, vector<uint32_t>> index_;
};

struct BlockInfo_ {
  uint32_t block_id;
  float grid_resolution;
  vector<uint8_t> bit_map;
  vector<uint32_t> active_grid_feature_index;
  vector<uint32_t> whole_grid_feature_index;
};

/////////// AbstractParser /////////////

enum LidarType
{
    HesaiP40 = 0,
    HesaiXT32
};

/////////// Trajectory /////////////

#include <fstream>

struct PPK_Raw_Info{
    uint32_t gps_week;
    double gps_sec, lat, lng, height,
    roll, pitch, yaw;
};

struct NODE_Raw_Info{
    uint32_t node_id;
    double time, x, y, z, w, qx, qy, qz;
};

enum DevicePosition{
    TopLidar = 0,
    BackLidar = 1,
    IMU = 2
};

struct TrajPoint{
    double timestamp = 0;
    uint32_t nodeId = 0;
    Vector3d translation = Vector3d::Identity();
    Quaterniond rotation = Quaterniond::Identity();
    EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

enum TrajType{
    PPK,
    NODE
};

/////////// MapTask /////////////

struct Time_Duration{
    double start_sec;
    double end_sec;
};

/////////// StrategyControl /////////////

enum Pointcloud_Source_Mode{
    PCAP_MODE,
    BAG_MODE,
    PCAP_BAG_MODE
};

enum PointCloudSource{
    PCAP,
    BAG
};

enum Trajectory_Source_Mode{
    NODE_MODE,
    PPK_MODE,
    PPK_UNDISORT_NODE_MAP_MODE
};

enum Filter_Mode{
    DYNAMIC_REMOVE_MODE,
    DIRECT_OUTPUT_MODE
};

struct StageInfo{
    DevicePosition lidarPosition;
    PointCloudSource pointcloudSource;
};

enum Strategy_Mode{
    MESH,
    PERIOD
};

enum Bag_Mode{
    Disorted,
    Undisorted
};

/////////// TaskServer /////////////

#define NODES_PER_SUBMAP (64)
#define SUBMAP_PACE (32)


struct TaskDescribe{
    string task_prefix;
    vector<vector<Time_Duration>> time_durations_vec;
    vector<string> output_prefix_vec;
};

/////////// DynamicRemover /////////////

const double MISS_PER_HIT_PER_STAGE = 4;


#endif // BASIC_TYPES_H
